#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Zach Way (zway1@gsu.edu)
# @Date: 2024-04-23
# @Filename: create_catalogidx_to_catalogidy.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import os
import time
from itertools import combinations

import numpy as np
import peewee
import yaml
from peewee import JOIN, fn
from playhouse.postgres_ext import ArrayField

from sdssdb.peewee.sdss5db.catalogdb import Catalog, database
from sdssdb.peewee.sdss5db.targetdb import Target

import target_selection


class TempMatch(peewee.Model):
    """Model for the first output table with all the matches."""

    pk = peewee.PrimaryKeyField()
    lead = peewee.TextField(null=False)
    catalogidx = peewee.BigIntegerField(index=True, null=False)
    catalogidy = peewee.BigIntegerField(index=True, null=False)
    version_idx = peewee.SmallIntegerField(index=True, null=False)
    version_idy = peewee.SmallIntegerField(index=True, null=False)

    class Meta:
        database = database
        schema = "sandbox"


class UniqueMatch(peewee.Model):
    """Model for the final table with unique pairs of catalogids"""

    pk = peewee.PrimaryKeyField()
    catalogidx = peewee.BigIntegerField(index=True, null=False)
    catalogidy = peewee.BigIntegerField(index=True, null=False)
    version_idx = peewee.SmallIntegerField(index=True, null=False)
    version_idy = peewee.SmallIntegerField(index=True, null=False)
    match_tables = ArrayField(peewee.TextField, null=False, index=False)
    n_tables = peewee.SmallIntegerField(null=False)

    class Meta:
        database = database
        schema = "sandbox"


class MetaXMatch:
    """Correlates crossmatch runs.

    This class correlates the catalogdb crossmatch runs by taking all the catalogids from a
    specified list (either with a table or in the config) and searches for all of the
    overlapping matches between versions. To do this it searches all
    the catalogids linked to the targetid associated to the target catalogid.

    To instantiate the object we need either the name of the configarion file or as
    configuration dictionary.

    The configuration file/dict contains the xmatch plans to be matched, the name of the log file,
    split_insert_number to indicate how many entries are ingested simultaneously to the first
    output table, the name of another configuration file indicating details about
    each crossmatch run, and 7 optional parameters. These parameters are sample_region to test
    the code in a single region, database options, show first to display the first N results
    in the log for each table, split_query to indicate the tables in which the main query
    is split into sub queries restricted to target_id ranges, ra_region to test the code in a
    range of right ascensions, individual_table to limit the code to a specific set of catalogids
    in the database, and catalogid_list to limit the code to a set of provided catalogids.
    By default, the code will run on all of Target. The individual crossmatch run
    config file indicates the version_id and tables of each xmatch run,
    mainly to calculate the intersecting tables between the 2 runs, to look for the matches
    in those tables.

    Parameters
    ----------
    database : peewee:PostgresqlDatabase
        A `peewee:PostgresqlDatabase` to the database the tables to
        cross-match.
    config_filename : str
        Configuration file containing at least parameters xmatch_plans, individual_xmatch_config
        log_file, and split_inrest_number.
    from_yaml : bool
        Load config from a yaml file? If True, parameter config_filename must also be specified.
    from_dict : bool
        Load config from a dictionary? If True, config_dict must be specified
    config_dict : dict
        Configuration dict containing at least parameters xmatch_plans, individual_xmatch_config
        log_file, and split_inrest_number.
    save_log_output : bool
        Save output to a file? Default is False.
    outer_join_sdss_id : bool
        Only use catalogids not associated with an sdss_id? If True, a match will be done to only
        run the catalogids through MetaXMatch that are not already in the sdss_id tables
    """

    def __init__(
        self,
        database,
        config_filename=None,
        from_yaml=True,
        from_dict=False,
        config_dict=None,
        save_log_output=False,
        outer_join_sdss_id=True,
    ):
        self.database = database
        self.outer_join_sdss_id = outer_join_sdss_id
        if from_yaml:
            config = yaml.load(open(config_filename, "r"), Loader=yaml.SafeLoader)
        elif from_dict:
            if type(config_dict) is not dict:
                raise ValueError("config must be a dict")
            config = config_dict

        log_filename = config["log_file"]
        self.log = target_selection.log
        if save_log_output:
            log_path_and_name = os.path.realpath("./" + log_filename)
            self.log.start_file_logger(log_path_and_name, rotating=False, mode="a")
            self.log.sh.setLevel(0)

        ind_xmatch_info_filename = config["individual_xmatch_config"]
        ind_xmatch_config = yaml.load(open(ind_xmatch_info_filename, "r"), Loader=yaml.SafeLoader)
        self.version_ids_to_match = config["version_ids_to_match"]

        optional_parameters = [
            "sample_region",
            "database_options",
            "show_first",
            "split_query",
            "ra_region",
            "individual_table",
            "catalogid_list",
        ]

        # All the optional parameters present in the configuration file are stored directly
        # As attributes of the MetaXMatch object and for the ones that are not the attribute
        # get a None value.

        for param in optional_parameters:
            if param in config.keys():
                setattr(self, param, config[param])
            else:
                setattr(self, param, None)

        # If sample_region is included in the configuration file, then the word full is ommited
        # from the output table and instead we include the ra and dec centers and the radius used

        if self.sample_region:
            racen, deccen, radius = self.sample_region
            st_dec = str(deccen)
            if "-" not in st_dec:
                st_dec = "pos" + st_dec
            else:
                st_dec = st_dec.replace("-", "neg")
            output_name = f"catalogidx_to_catalogidy" f"_ra{racen}_dec{st_dec}_rad{int(radius)}"
            log_message = f"###  Using ra={racen:5.1f}  dec={deccen:5.1f}  rad={radius:04.1f}  ###"
        elif self.ra_region:
            ra_start, ra_stop = self.ra_region
            if ra_start >= ra_stop:
                raise ValueError("Starting RA must be smaller than stopping RA!")
            output_name = f"catalogidx_to_catalogidy" f"_ra{int(ra_start)}_ra{int(ra_stop)}"
            log_message = f"###  Using ra={ra_start:5.1f}  to  ra={ra_stop:5.1f}               ###"
        elif self.individual_table:
            ind_table = self.individual_table
            if "." in ind_table:
                ind_table = ind_table.split(".")[-1]
            output_name = f"catalogidx_to_catalogidy" f"_{ind_table}"
            log_message = f"###  Using catalogids from file  {ind_table}              ###"
        elif self.catalogid_list:
            output_name = "catalogidx_to_catalogidy_from_catalogid_list"
            log_message = "### Using catalogids from a given list                    ###"
        else:
            output_name = "catalogidx_to_catalogidy_all"
            log_message = "###            Using Full SKY             ###"

        if self.ra_region and self.individual_table:
            ra_start, ra_stop = self.ra_region
            ind_table = self.individual_table
            if "." in ind_table:
                ind_table = ind_table.split(".")[-1]
            output_name = (
                f"catalogidx_to_catalogidy" f"_{ind_table}" f"_ra{int(ra_start)}_ra{int(ra_stop)}"
            )
            log_message = (
                f"###  Using catalogids from file  {ind_table} "
                + f" in ra={ra_start:5.1f} to ra={ra_stop:5.1f} ###"
            )

        self.output_name = output_name
        self.log.info(" ")
        self.log.info("#" * 45)
        self.log.info(log_message)
        self.log.info("#" * 45)

        all_tables = [set(ind_xmatch_config[k]["order"]) for k in ind_xmatch_config.keys()]
        intersecting_tables = set()
        for i_comb in list(combinations([i for i in range(len(all_tables))], 2)):
            intersecting_tables = intersecting_tables | (
                all_tables[i_comb[0]] & all_tables[i_comb[1]]
            )

        intersecting_tables = list(intersecting_tables)
        intersecting_tables.sort()
        self.intersecting_tables = intersecting_tables
        self.log.info(f"The intersecting tables to match are: {intersecting_tables}")
        self.split_insert_nunmber = config["split_insert_nunmber"]

        if self.split_query:
            self.split_query_info = {el[0]: (el[1], el[2]) for el in self.split_query}

    def match_in_table(self, name, split=False, min_targetid=0, max_targetid=0):
        """This method takes as argument the name of the table to be matched to
        do the query and find the catalogid pairs. This method can also be used
        with argument split=True in which case argument min_targetid, and max_targetid are used
        to do a sub query restricted to a target_id range. For each input table ingested
        in both crossmatch runs the main query starts by creating a cte querying the target
        table and storing the target_id in the relational table associated with the catalogids
        from the target table that belong to the older crossmatch run version_id.
        Then the other query takes the target_ids from the cte and checks which catalogid's from
        both versions are associated with that target_id. Then for each target_id from the query
        it looks for all the possible pairs of catalogids in which one catalogid comes from the
        first crossmatch run and the other comes from the second crossmatch run.
        Finally these catalogid pairs for each input table are ingested in a table with suffix all

        Parameters
        ----------
        name : str
            Name of the table crossmatched in both runs in which the catalogid pais will be looked
        split : bool
            If True the query is split into multiple subqueries restricted to ranges of targetid
            To use this for a specific table the configuration file parameter split query must
            contain as an element a list where the first element is the name of the table,
            the second one is a number larger than the maximum targetid of the table,
            and the third one is the target id bin width to be used to split the queries
        min_targetid : int
            minimum target_id to be used for this specific query
        max_targetid : int
            maximum target_id to be used for this specific query

        """

        # First we set the database options for the query
        if self.database_options:
            for param in self.database_options:
                value = self.database_options[param]
                setting_message = f"SET {param}={value!r};"
                self.database.execute_sql(setting_message)
                self.log.info(setting_message)

        temp_fields = [
            TempMatch.lead,
            TempMatch.catalogidx,
            TempMatch.catalogidy,
            TempMatch.version_idx,
            TempMatch.version_idy,
        ]

        db_models = self.database.models

        tables = [db_models["catalogdb." + name] for name in self.intersecting_tables]
        rel_tables = [
            db_models["catalogdb.catalog_to_" + name] for name in self.intersecting_tables
        ]

        index = self.intersecting_tables.index(name)
        table = tables[index]
        rel_table = rel_tables[index]
        tablename = table._meta.table_name

        if self.individual_table:
            try:
                ind_table = db_models[self.individual_table]
            except:
                raise ValueError(
                    "Could not find the table in database model: " + self.individual_table
                )

            if self.individual_table == "catalogdb.catalog_to_" + name:
                cte_targetids = (
                    Catalog.select(rel_table.target_id, Catalog.ra, Catalog.dec)
                    .join(rel_table, on=(Catalog.catalogid == rel_table.catalogid))
                    .where((rel_table.version << self.version_ids_to_match) & (rel_table.best))
                    .distinct(rel_table.target_id)
                )

            else:
                cte_targetids = (
                    Catalog.select(rel_table.target_id, Catalog.ra, Catalog.dec)
                    .join(ind_table, on=(Catalog.catalogid == ind_table.catalogid))
                    .join(rel_table, on=(Catalog.catalogid == rel_table.catalogid))
                    .where((rel_table.version << self.version_ids_to_match) & (rel_table.best))
                    .distinct(rel_table.target_id)
                )

        elif self.catalogid_list:
            cte_targetids = (
                Catalog.select(rel_table.target_id, Catalog.ra, Catalog.dec)
                .join(rel_table, on=(Catalog.catalogid == rel_table.catalogid))
                .where(((rel_table.version << self.version_ids_to_match) & (rel_table.best)))
                .where(Catalog.catalogid << self.catalogid_list)
                .distinct(rel_table.target_id)
            )

        else:
            cte_targetids = (
                Target.select(rel_table.target_id, Target.ra, Target.dec)
                .join(rel_table, on=(Target.catalogid == rel_table.catalogid))
                .where((rel_table.version << self.version_ids_to_match) & (rel_table.best))
                .distinct(rel_table.target_id)
            )

        # If sample_region or ra_region is included in the configuration file,
        # then the entire process is restricted to that region

        if self.sample_region:
            racen, deccen, radius = self.sample_region
            cte_targetids = cte_targetids.where(
                fn.q3c_radial_query(Target.ra, Target.dec, racen, deccen, radius)
            )
        elif self.ra_region and self.individual_table:
            ra_start, ra_stop = self.ra_region
            cte_targetids = cte_targetids.where(Catalog.ra.between(ra_start, ra_stop))
        elif self.ra_region:
            ra_start, ra_stop = self.ra_region
            cte_targetids = cte_targetids.where(Target.ra.between(ra_start, ra_stop))

        # If outer_join_sdss_id is True, only match in the catalogids that aren't
        # already in sdss_id_flat

        if self.outer_join_sdss_id:
            cte_targetids = cte_targetids.join(
                target_selection.sdss_id.SdssIdFlat,
                join_type=JOIN.LEFT_OUTER,
                on=(rel_table.catalogid == target_selection.sdss_id.SdssIdFlat.catalogid),
            ).where(target_selection.sdss_id.SdssIdFlat.catalogid.is_null())

        cte_targetids = cte_targetids.cte("cte_targets")

        query = (
            rel_table.select(
                rel_table.target_id,
                peewee.Value(tablename).alias("table"),
                fn.array_agg(rel_table.catalogid)
                .over(partition_by=rel_table.target_id)
                .alias("catalogids"),
                fn.array_agg(rel_table.version_id)
                .over(partition_by=rel_table.target_id)
                .alias("version_ids"),
            )
            .join(cte_targetids, on=(rel_table.target_id == cte_targetids.c.target_id))
            .where((rel_table.version << self.version_ids_to_match) & (rel_table.best))
            .distinct(rel_table.target_id)
            .with_cte(cte_targetids)
        )

        if split:
            query = query.where(
                rel_table.target_id >= min_targetid, rel_table.target_id < max_targetid
            )

        if not split:
            self.log.info(" ")
            self.log.info(" ")

        # of catalogids associated to it where one comes from the first crossmatch run and
        # the other one from the second crossmatch run

        res = query.dicts()
        results_list = []
        npair = 0
        nres = 0
        for entry in res:
            nres += 1
            catids = entry["catalogids"]
            verids = entry["version_ids"]
            catid_combs = list(combinations(catids, 2))  # Check all the pairs of catalogid's
            verid_combs = list(combinations(verids, 2))  # Check all the pairs of version_id's
            ind_valid = [
                ind
                for ind in range(len(catid_combs))
                if catid_combs[ind][0] != catid_combs[ind][1]
            ]
            for ind_comb in ind_valid:  # For each pair comming from different version_id's
                npair += 1
                sorted_catids = np.sort(catid_combs[ind_comb])
                sorted_verids = np.sort(verid_combs[ind_comb])
                curr_tuple = (
                    entry["table"],
                    sorted_catids[0],
                    sorted_catids[1],
                    sorted_verids[0],
                    sorted_verids[1],
                )
                results_list.append(curr_tuple)
                if self.show_first and npair <= self.show_first and not split:
                    tuple_log = str(npair) + "  " + "  ".join([str(el) for el in curr_tuple])
                    self.log.info(tuple_log)

        # In this final part we go over the results and split it in chunks of N entries
        # each time ingesting those entries to the output table

        parts = int((len(results_list) - 1) / self.split_insert_nunmber) + 1

        # Check if connection has timed out. Reconnect if it has
        database.connect(
            dbname=database.dbname,
            user=database.connection_params.get("user") if database.connection_params else None,
            reuse_if_open=True,
        )

        for npart in range(parts):
            with self.database.atomic():
                if len(results_list) == 0:
                    continue
                if not split:
                    self.log.info(f"Ingesting part {npart + 1} out of {parts}")
                first_el = npart * self.split_insert_nunmber
                last_el = (npart + 1) * self.split_insert_nunmber
                curr_results = results_list[first_el:last_el]
                TempMatch.insert_many(curr_results, fields=temp_fields).execute()

        return nres, npair

    def run(self, starting_table=None):
        """Loops over the input tables to check the catalogid pairs associated to each target id.

        This method sets the models creates the tables and then goes over all the tables
        to look in each one the catalogid matches from both runs. It can be used with
        the starting_table name option to skip all the tables before that one in the process.

        Parameters
        ----------
        starting_table : str
            Name of the table in which the process would be restarted. By default tables are
            ordered alphabetically.
        """

        t0 = time.time()
        tables_to_process = self.intersecting_tables
        if starting_table:
            ind = tables_to_process.index(starting_table)
            tables_to_process = tables_to_process[ind:]

        TempMatch._meta.table_name = self.output_name
        self.database.bind([TempMatch])
        if TempMatch.table_exists():
            self.database.drop_tables([TempMatch])
            self.log.info(f"Dropped table {TempMatch._meta.table_name}")

        # self.database.create_tables([TempMatch])
        TempMatch.create_table()
        self.log.info(f"Created table {TempMatch._meta.table_name}")

        add_index_TempMatch = f"""CREATE INDEX ON
                                    sandbox.{self.output_name} (catalogidx);
                                  CREATE INDEX ON
                                    sandbox.{self.output_name} (catalogidy);
                                  CREATE INDEX ON
                                    sandbox.{self.output_name} (version_idx);
                                  CREATE INDEX ON
                                    sandbox.{self.output_name} (version_idy);"""
        self.database.execute_sql(add_index_TempMatch)

        for curr_table in tables_to_process:
            t1 = time.time()
            logname = "Running table " + curr_table
            self.log.info("-" * 52)
            self.log.info(f"|{logname:^50}|")
            self.log.info("-" * 52)

            if self.split_query and curr_table in self.split_query_info.keys():
                ntargetids, npairs = 0, 0
                self.log.info(f"Doing the query of {curr_table} splitting by targetid")
                max_tar, bin_tar = self.split_query_info[curr_table]
                self.log.info(f"Using binning={bin_tar} and max value ={max_tar}")
                min_targetid, max_targetid = 0, bin_tar
                while min_targetid < max_tar:
                    ti_iter = time.time()
                    curr_targetids, curr_pairs = self.match_in_table(
                        curr_table,
                        split=True,
                        min_targetid=min_targetid,
                        max_targetid=max_targetid,
                    )
                    ntargetids += curr_targetids
                    npairs += curr_pairs
                    min_targetid += bin_tar
                    max_targetid += bin_tar
                    tf_iter = time.time()
                    self.log.info(
                        f"Ingested {curr_pairs} to output table using targetid range "
                        f"{min_targetid} to {max_targetid} "
                        f"in {(tf_iter - ti_iter):.2f} seconds"
                    )
            else:
                ntargetids, npairs = self.match_in_table(curr_table)
            t2 = time.time()
            self.log.info(f"Found {ntargetids} targetids with multiple catalogids associated")
            self.log.info(f"{npairs} valid catalogid pairs were ingested to output table ")
            self.log.info(f"Table {curr_table} was processed in {(t2 - t1):.2f} seconds")
        t3 = time.time()
        self.log.info(f"The entire match took {(t3 - t0):.2f} seconds")


def create_unique_from_region(input_tablename, save_log_output=False):
    """Function to create the final output table with unlist of unique pairs of catalogid.
    For a regional match.

    This function takes the already created table with all the catalogid pairs for each input
    table and selects the list of unique pairs indicating for each pair the table in which it
    was found and the number of tables in which it was found

    Parameters
    ----------
    input_tablename : string
        Table name of crossmatch run
    save_log_output : bool
        Save the output log to a file?
    """
    ti = time.time()
    output_tablename = input_tablename + "_unique"
    UniqueMatch._meta.table_name = output_tablename
    database.bind([UniqueMatch])
    log = target_selection.log
    if save_log_output:
        log_path_and_name = os.path.realpath("./" + input_tablename + "_unique_table_creation.log")
        log.start_file_logger(log_path_and_name, rotating=False, mode="a")

    if UniqueMatch.table_exists():
        database.drop_tables([UniqueMatch])
        log.info(f"Dropped table {output_tablename}")
    # database.create_tables([UniqueMatch])
    UniqueMatch.create_table()
    add_index_UniqueMatch = f"""CREATE INDEX ON
                                    sandbox.{output_tablename} (catalogidx);
                                CREATE INDEX ON
                                    sandbox.{output_tablename} (catalogidy);
                                CREATE INDEX ON
                                    sandbox.{output_tablename} (version_idx);
                                CREATE INDEX ON
                                    sandbox.{output_tablename} (version_idy);"""
    database.execute_sql(add_index_UniqueMatch)
    log.info(f"Created table {output_tablename}")
    TempMatch._meta.table_name = input_tablename
    query = TempMatch.select(
        TempMatch.catalogidx,
        TempMatch.catalogidy,
        TempMatch.version_idx,
        TempMatch.version_idy,
        fn.array_agg(TempMatch.lead),
        fn.count(TempMatch.lead),
    ).group_by(
        TempMatch.catalogidx, TempMatch.catalogidy, TempMatch.version_idx, TempMatch.version_idy
    )

    fields = [
        UniqueMatch.catalogidx,
        UniqueMatch.catalogidy,
        UniqueMatch.version_idx,
        UniqueMatch.version_idy,
        UniqueMatch.match_tables,
        UniqueMatch.n_tables,
    ]

    insert_query = UniqueMatch.insert_from(query, fields).returning()
    cursor = insert_query.execute()
    n_unique = cursor.rowcount
    tf = time.time()
    log.info(f"Created unique pairs table with {n_unique} " f"entries in {(tf - ti):.2f} seconds")


"""
EXAMPLE OF USAGE
metax = MetaXMatch('cat_to_cat.yml', database)
metax.run()
create_unique_table("output_name")
Where "output_name" is the catalogidx_to_catalogidy_* output name from metax.run()
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2020-12-14
# @Filename: ops_boss_stds.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

# additional imports required by ops_boss_stds_ps1dr2
from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToLegacy_Survey_DR8,
    CatalogToLegacy_Survey_DR10,
    CatalogToPanstarrs1,
    CatalogToSDSS_DR13_PhotoObj_Primary,
    CatalogToTIC_v8,
    Gaia_DR2,
    Legacy_Survey_DR8,
    Legacy_Survey_DR10,
    Panstarrs1,
    TIC_v8,
    TwoMassPSC,
    eBOSS_Target_v5,
)

from target_selection.cartons import BaseCarton
from target_selection.mag_flux import AB2Jy


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py
#
# peewee Model name ---> postgres table name
# Gaia_DR2(CatalogdbModel)--->'gaia_dr2_source'
# TwoMassPSC(CatalogdbModel) --->'twomass_psc'
# eBOSS_Target_v5(CatalogdbModel)--->'ebosstarget_v5'
# Legacy_Survey_DR8(CatalogdbModel)--->'legacy_survey_dr8'
#
# The CatalogTo* peewee model names are not explicit in catalogdb.py.
# The names are constructed by prepending CatalogTo
# to the corresponding model name which is in catalogdb.py.
# For example:
# CatalogToSDSS_DR13_PhotoObj_Primary--->'catalog_to_sdss_dr13_photoobj_primary'
# CatalogToLegacy_Survey_DR8--->'catalogdb.catalog_to_legacy_survey_dr8'
#
# In the carton code, peewee.fn.log() is calling
# the PostgreSQL log() which is a base 10 logarithm.
# The below link has more details:
# https://www.postgresql.org/docs/12/functions-math.html


class OPS_BOSS_Stds_Carton(BaseCarton):
    """
    Shorthand name: ops_boss_stds
    Selection criteria:
    --- criteria for ops_BOSS_stds ---
    #calculate distance modulus (could convert to use BailerJones distance)
    distMod = 5.*np.log10(1000./gaia_DR2.parallax)-5.

    #calculate absolute g and k magnitudes
    abs_gmag = gaia_DR2.phot_g_mean_mag - distMod
    abs_kmag = twomass_psc.k_m - distMod

    meet_std_criteria =
    np.where( ( ( (gaia_DR2.phot_bp_mean_mag -
                   gaia_DR2.phot_rp_mean_mag) >= 0.65) &
                  ( (gaia_DR2.phot_bp_mean_mag -
                     gaia_DR2.phot_rp_mean_mag) <= 0.8) &
                  ( (abs_gmag) >= 3.5) & ( (abs_gmag) <= 5.5) )
             )
    Also add the below condition:
             Gaia_DR2.phot_g_mean_mag > 13 AND
             Gaia_DR2.phot_g_mean_mag < 18.5

    Lead contact: Kevin Covey
    """

    name = "ops_std_boss"
    category = "standard_boss"
    cadence = None
    program = "ops_std"
    priority = 5425
    mapper = None
    instrument = "BOSS"
    can_offset = False

    def build_query(self, version_id, query_region=None):
        distMod = 5.0 * peewee.fn.log(1000.0 / Gaia_DR2.parallax) - 5.0
        abs_gmag = Gaia_DR2.phot_g_mean_mag - distMod

        # We need a join with Catalog because
        # when using coordinates in the cartons we want to always
        # use Catalog.ra and Catalog.dec.
        # This is because those coordinates have
        # all been put in a common epoch 2015.5.
        # Below, we are using Catalog.ra and Catalog.dec
        # inside the "if query_region" block.

        query = (
            Catalog.select(CatalogToTIC_v8.catalogid)
            .join(CatalogToTIC_v8, on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                Gaia_DR2.parallax > 0,
                (Gaia_DR2.phot_bp_mean_mag - Gaia_DR2.phot_rp_mean_mag) >= 0.65,
                (Gaia_DR2.phot_bp_mean_mag - Gaia_DR2.phot_rp_mean_mag) <= 0.8,
                abs_gmag >= 3.5,
                abs_gmag <= 5.5,
                Gaia_DR2.phot_g_mean_mag > 13,
                Gaia_DR2.phot_g_mean_mag < 18.5,
            )
        )
        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            # Here we do not need a join with Catalog since query already contains
            # a join with Catalog above.
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query


class OPS_BOSS_Red_Stds_Deredden_Carton(BaseCarton):
    """
        Shorthand name: ops_boss_red_stds_deredden

        Selection Criteria:
        This carton has selection criteria as defined in the below SQL code.

        SELECT c.catalogid, c.ra, c.dec, c.pmra, c.pmdec,
                twomass.j_m, twomass.h_m, twomass.k_m,
                gaia.phot_g_mean_mag, gaia.phot_bp_mean_mag, gaia.phot_rp_mean_mag,
                 gaia.parallax, gaia.pmra, gaia.pmdec

        FROM catalog c

        INNER JOIN catalog_to_tic_v8 ctic USING (catalogid)

        INNER JOIN tic_v8 tic ON tic.id = ctic.target_id

        INNER JOIN gaia_dr2_source gaia ON gaia.source_id = tic.gaia_int

        INNER JOIN twomass_psc twomass ON twomass.designation = tic.twomass_psc

        WHERE gaia.parallax > 0 AND gaia.phot_g_mean_mag < 18 AND

        /* enforce bp - rp edges of reddening strip in bp-rp vs. M_G space */

        gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag > 0.95 AND

        gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag < 3.625 AND

        /* enforce top and bottom of reddening strip in bp-rp vs. M_G space */

        gaia.phot_g_mean_mag - 5.*log(1000./gaia.parallax) + 5. >
         1.15+(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag)*1.89 AND

        gaia.phot_g_mean_mag - 5.*log(1000./gaia.parallax) + 5. <
         3.4+(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag)*1.89 AND

        /* enforce limits in empirical g-k color (make sure the source is at
         least a little reddened relative to a standard F star) */

        gaia.phot_g_mean_mag - twomass.k_m > 2.5 AND

        /* enforce limits in dereddened g-k color */

        (gaia.phot_g_mean_mag - 1.89*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) )
         - (twomass.k_m-0.186*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) ) > 1.2 AND

        (gaia.phot_g_mean_mag - 1.89*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) )
         - (twomass.k_m-0.186*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) ) < 1.8 AND

        /* enforce limits in dereddened j-k color */

        (twomass.j_m-0.582*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) )
         - (twomass.k_m-0.186*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) ) > 0.15 AND

        (twomass.j_m-0.582*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) )
         - (twomass.k_m-0.186*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) ) < 0.45 AND

        /* enforce limits in dereddened absolute k magnitude */

        twomass.k_m - 5.*log(1000./gaia.parallax) + 5. -
        0.186*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) > 2 AND

        twomass.k_m - 5.*log(1000./gaia.parallax) + 5. -
         0.186*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag - 0.7125) < 3.25 AND

        /* enforce proper motion limit */

        SQRT(gaia.pmra^2 + gaia.pmdec^2) > 3.5 AND

        (gaia.phot_rp_mean_mag + 5.*log(SQRT(gaia.pmra^2 + gaia.pmdec^2))-10.)
         > 10.11-1.43*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag) AND

        (gaia.phot_rp_mean_mag + 5.*log(SQRT(gaia.pmra^2 + gaia.pmdec^2))-10.)
         < 13.11-1.43*(gaia.phot_bp_mean_mag - gaia.phot_rp_mean_mag)
    ;
        Lead contact: Kevin Covey
    """

    name = "ops_std_boss_red"
    category = "standard_boss"
    cadence = None
    program = "ops_std"
    priority = 5450
    mapper = None
    instrument = "BOSS"
    can_offset = False

    def build_query(self, version_id, query_region=None):
        # calculate proper motion amplitude and
        # term to construct reduced proper motion diagram
        pm_amp = peewee.fn.sqrt(Gaia_DR2.pmra * Gaia_DR2.pmra + Gaia_DR2.pmdec * Gaia_DR2.pmdec)
        pm_Mod = 5.0 * peewee.fn.log(pm_amp) - 10.0

        # calculate a parallax-based distance modulus
        distMod = 5.0 * peewee.fn.log(1000.0 / Gaia_DR2.parallax) - 5.0

        # calculate the absolute g magnitude
        abs_gmag = Gaia_DR2.phot_g_mean_mag - distMod
        abs_kmag = TwoMassPSC.k_m - distMod

        # infer reddenings associated with assumed Fstar colors
        fstar_bp_rp = 0.725
        ag = 1.890 * (Gaia_DR2.bp_rp - fstar_bp_rp)
        aj = 0.582 * (Gaia_DR2.bp_rp - fstar_bp_rp)
        ak = 0.186 * (Gaia_DR2.bp_rp - fstar_bp_rp)

        query = (
            Catalog.select(
                CatalogToTIC_v8.catalogid,
                Catalog.ra,
                Catalog.dec,
                Gaia_DR2.source_id,
                Gaia_DR2.phot_g_mean_mag,
                Gaia_DR2.phot_bp_mean_mag,
                Gaia_DR2.phot_rp_mean_mag,
                TwoMassPSC.j_m,
                TwoMassPSC.k_m,
                ag.alias("ag"),
                ak.alias("ak"),
            )
            .join(CatalogToTIC_v8, on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .switch(TIC_v8)
            .join(TwoMassPSC, on=(TIC_v8.twomass_psc == TwoMassPSC.designation))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                # require a parallax so that absolute magnitudes are meaningful
                Gaia_DR2.parallax > 0,
                # require a slightly non-zero pm amplitude to
                # a) ensure reduced proper motion is meaningful, and
                # b) modestly filter quasars as candidates
                pm_amp > 3.5,
                # require Gaia G < 18 to provide reasonable SNR, and
                # G-K > 2 to ensure measured colors are at least
                # slightly reddened w.r.t regular f stars
                Gaia_DR2.phot_g_mean_mag < 18,
                Gaia_DR2.phot_g_mean_mag - TwoMassPSC.k_m > 2.5,
                # enforce bp - rp edges of reddening strip in
                # bp-rp vs. M_G space
                Gaia_DR2.phot_bp_mean_mag - Gaia_DR2.phot_rp_mean_mag > 0.95,
                Gaia_DR2.phot_bp_mean_mag - Gaia_DR2.phot_rp_mean_mag < 3.625,
                # enforce top and bottom of reddening strip
                # in bp-rp vs. M_G space
                abs_gmag > (1.15 + ag),
                abs_gmag < (3.4 + ag),
                # enforce limits in dereddened g-k color
                (Gaia_DR2.phot_g_mean_mag - ag) - (TwoMassPSC.k_m - ak) > 1.2,
                (Gaia_DR2.phot_g_mean_mag - ag) - (TwoMassPSC.k_m - ak) < 1.8,
                # enforce limits in dereddened j-k color
                (TwoMassPSC.j_m - aj) - (TwoMassPSC.k_m - ak) > 0.15,
                (TwoMassPSC.j_m - aj) - (TwoMassPSC.k_m - ak) < 0.45,
                # enforce limits in dereddened absolute k magnitude
                abs_kmag - ak > 2,
                abs_kmag - ak < 3.25,
                # enforce halo-like motions in the reduced proper motion
                # diagram (selected in a strip consistent
                # w/ reddening the existing eBOSS standards)
                (Gaia_DR2.phot_rp_mean_mag + pm_Mod)
                > (10.11 - 1.43 * (Gaia_DR2.phot_bp_mean_mag - Gaia_DR2.phot_rp_mean_mag)),
                (Gaia_DR2.phot_rp_mean_mag + pm_Mod)
                < (13.11 - 1.43 * (Gaia_DR2.phot_bp_mean_mag - Gaia_DR2.phot_rp_mean_mag)),
            )
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query


class OPS_eBOSS_Stds_Carton(BaseCarton):
    """
    Shorthand name: ops_eboss_stds
    Selection Criteria:
    The code of this carton is based on the below SQL.
    SELECT DISTINCT e.objid_targeting
       FROM ebosstarget_v5 e JOIN catalog_to_sdss_dr13_photoobj_primary s
       ON s.target_id = e.objid_targeting
       WHERE (e.eboss_target1 & pow(2, 50)::bigint) > 0 OR
             (e.eboss_target1 & pow(2, 51)::bigint) > 0 OR
             (e.eboss_target1 & pow(2, 52)::bigint) > 0
             and s.best is true and s.version_id = 21;

    Lead contact: Kevin Covey
    """

    name = "ops_std_eboss"
    category = "standard_boss"
    cadence = None
    program = "ops_std"
    priority = 5300
    mapper = None
    instrument = "BOSS"
    can_offset = False

    def build_query(self, version_id, query_region=None):
        selection_condition = (
            (eBOSS_Target_v5.eboss_target1.bin_and(peewee.fn.pow(2, 50).cast("bigint")) > 0)
            | (eBOSS_Target_v5.eboss_target1.bin_and(peewee.fn.pow(2, 51).cast("bigint")) > 0)
            | (eBOSS_Target_v5.eboss_target1.bin_and(peewee.fn.pow(2, 52).cast("bigint")) > 0)
        )

        # We have distinct(eBOSS_Target_v5.objid_targeting) at the end
        # since the table ebosstarget_v5 has duplicate values.
        query = (
            Catalog.select(Catalog.catalogid)
            .join(
                CatalogToSDSS_DR13_PhotoObj_Primary,
                on=(Catalog.catalogid == CatalogToSDSS_DR13_PhotoObj_Primary.catalogid),
            )
            .join(
                eBOSS_Target_v5,
                on=(
                    CatalogToSDSS_DR13_PhotoObj_Primary.target_id
                    == eBOSS_Target_v5.objid_targeting
                ),
            )
            .where(
                CatalogToSDSS_DR13_PhotoObj_Primary.version_id == version_id,
                CatalogToSDSS_DR13_PhotoObj_Primary.best >> True,
                selection_condition,
            )
            .distinct(eBOSS_Target_v5.objid_targeting)
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query


class OPS_BOSS_Stds_TIC_Carton(BaseCarton):
    """
    Shorthand name: ops_boss_stds_tic

    Simplified Description of selection criteria: parent catalog is TIC.

    13 < G < 17   (Gaia mag between 13 and 17)
    6000 < TIC teff < 8000     (temp between 6000-8000 K, so in
     late A - F - early G spectral range)
    3 < TIC log g < 5.5    (select to physically meaningful log g
     values to exclude nulls etc.)
    Once TIC has been limited to sources meeting the above criteria,
     the sky is then divided into an nside = 128 HEALPIX skymap.
    The 10 highest gravity sources in each healpix are then selected and
     saved for the output carton.

    All HEALPix use the nested ordering.

    Return columns: All the filter columns (Gaiamag, teff, logg),
    all other optical magnitudes in the
    TIC (bmag, vmag, umag, gmag, rmag, imag, zmag) plus healpix_128  and
    2MASS ID + H magnitude (if it exists, null if not;
    saved in case we need to use as fillers for high latitude BHM plates)

    Pseudo SQL (optional):

    Cadence options for these targets (list all options,
    even though no single target will receive more than one): N/A

    Lead contact:  Kevin Covey (but maybe also Marina Kounkel
    & Hector Javier Ibarra-Medel )
    """

    name = "ops_std_boss_tic"
    category = "standard_boss"
    cadence = None
    program = "ops_std"
    priority = 5400
    mapper = None
    instrument = "BOSS"
    can_offset = False

    def build_query(self, version_id, query_region=None):
        query = (
            Catalog.select(
                CatalogToTIC_v8.catalogid,
                Catalog.ra,
                Catalog.dec,
                TIC_v8.hmag,
                TIC_v8.twomass_psc,
                Gaia_DR2.phot_g_mean_mag,
                TIC_v8.teff,
                TIC_v8.logg.alias("logg"),
                TIC_v8.bmag,
                TIC_v8.vmag,
                TIC_v8.umag,
                TIC_v8.gmag,
                TIC_v8.rmag,
                TIC_v8.imag,
                TIC_v8.zmag,
                peewee.fn.healpix_ang2ipix_nest(128, Catalog.ra, Catalog.dec).alias("healpix_128"),
            )
            .join(CatalogToTIC_v8, on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .join(Gaia_DR2, on=(TIC_v8.gaia_int == Gaia_DR2.source_id))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                Gaia_DR2.phot_g_mean_mag > 13,
                Gaia_DR2.phot_g_mean_mag < 17,
                TIC_v8.teff > 6000,
                TIC_v8.teff < 8000,
                TIC_v8.logg > 3,
                TIC_v8.logg < 5.5,
            )
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )
        return query

    def post_process(self, model):
        """
        The 10 highest gravity sources in each healpix are then
        selected and saved for the output carton.in each healpix pixel.
        """

        self.database.execute_sql("update sandbox.temp_ops_std_boss_tic " + "set selected = false")

        cursor = self.database.execute_sql(
            "select catalogid, healpix_128, logg from "
            + " sandbox.temp_ops_std_boss_tic "
            + " order by healpix_128 asc, logg desc;"
        )

        output = cursor.fetchall()

        list_of_catalog_id = [0] * len(output)
        nside = 128
        total_number_healpix_pixels = 12 * nside * nside
        count = [0] * total_number_healpix_pixels
        current_target = 0
        for i in range(len(output)):
            current_healpix = output[i][1]
            if count[current_healpix] < 10:
                count[current_healpix] = count[current_healpix] + 1
                list_of_catalog_id[current_target] = output[i][0]
                current_target = current_target + 1

        max_target = current_target
        for k in range(max_target + 1):
            self.database.execute_sql(
                " update sandbox.temp_ops_std_boss_tic set selected = true "
                + " where catalogid = "
                + str(list_of_catalog_id[k])
                + ";"
            )


class OPS_BOSS_Stds_LSDR8_Carton(BaseCarton):
    """
    Shorthand name: ops_boss_stds_lsdr8
    – CRITERIA STILL IN VETTING STAGE -
    probably need to repeat all
    of the below in de-reddened colour space

    Comments: Spectrophotometric standards suitable for use by BOSS in dark time.

    Simplified Description of selection criteria:
    This simplified description is very high level.
    The carton is based on the pseudo SQL below.

    Parent catalog is legacy_survey_dr8 as ls, gaia_dr2_source as g2 with criteria:

    g2.G > 15.5
    15.95 < ls.r < 18.05
    -0.5 < g2.parallax < 1.0
    0.254 < ls.g - ls.r < 0.448
     0.024 < ls.r - ls.z < 0.190
    0.619 < g2.bp - g2.rp < 0.863
    0.0 < g2.G - ls.r < 0.10
    ls.type = 'PSF'
    ls.ref_cat = 'G2'
    ls.nobs_g > 2 && ls.nobs_r > = 2 && ls.nobs_z >= 2
    ls.maskbits = 0
    ls.gaia_duplicated_source = False
    Wiki page:
    All-sky BOSS standards#skyBOSSstandards-TransferringeBOSSselectiontolegacy_survey_dr8

    Return columns: TBD

    Pseudo SQL (optional):

    SELECT
        c.catalogid,l.ls_id,
        c.ra,c.dec,
        l.flux_g,l.flux_r,l.flux_z,l.flux_w1,
        l.flux_ivar_g,l.flux_ivar_r,l.flux_ivar_z,l.flux_ivar_w1,
        l.gaia_phot_g_mean_mag, l.gaia_phot_bp_mean_mag, l.gaia_phot_rp_mean_mag,
        l.parallax,l.parallax_ivar
    FROM catalog AS c
    JOIN catalog_to_legacy_survey_dr8 AS c2l
        ON c.catalogid = c2l.catalogid
    JOIN legacy_survey_dr8 AS l
        ON c2l.target_id = l.ls_id
    WHERE l.type = 'PSF'
        AND ref_cat = 'G2'
        AND gaia_phot_g_mean_mag > 15.5
        AND ((22.5 - 2.5*log10(greatest(1e-9,l.flux_r))) BETWEEN 15.95 AND 18.05)
        AND (l.parallax BETWEEN -0.5 AND 1.0)
        AND ((-2.5*log10(greatest(1e-9,l.flux_g)/
        greatest(1e-9,l.flux_r))) BETWEEN 0.254 AND 0.448)
        AND ((-2.5*log10(greatest(1e-9,l.flux_r)/
        greatest(1e-9,l.flux_z))) BETWEEN 0.024 AND 0.190)
        AND ((l.gaia_phot_bp_mean_mag-l.gaia_phot_rp_mean_mag)
         BETWEEN 0.619 AND 0.863)
        AND ((l.gaia_phot_g_mean_mag - (22.5-2.5*log10(greatest(1e-9,l.flux_r))))
         BETWEEN 0.0 AND 0.10)
        AND gaia_duplicated_source = false
        AND l.nobs_g >=2
        AND l.nobs_r >=2
        AND l.nobs_z >=2
        AND maskbits = 0
        AND c.version_id = ...
        AND c2l.best = true
    Cadence options for these targets (list all options,
     even though no single target will receive more than one): N/A

    Lead contact:  Tom Dwelly
    """

    name = "ops_std_boss_lsdr8"
    category = "standard_boss"
    cadence = None
    program = "ops_std"
    priority = 5350
    mapper = None
    instrument = "BOSS"
    can_offset = False

    def build_query(self, version_id, query_region=None):
        ls = Legacy_Survey_DR8.alias()
        c2ls = CatalogToLegacy_Survey_DR8.alias()

        # an alias to simplify accessing the query parameters:
        pars = self.parameters

        # safety catch to avoid log-of-zero and divide by zero errors.
        # => use a flux in nano-maggies below which we give up
        nMgy_min = 1e-3  # equiv to AB=30

        # Below line is used to avoid divide by zero or log of zero,
        #     peewee.fn.greatest(nMgy_min, Legacy_Survey_DR8.flux_g)
        # Below peewee.fn.log is log to the base 10.
        # peewee.fn.log(peewee.fn.greatest(nMgy_min, Legacy_Survey_DR8.flux_r))

        # transform the legacysurvey grz into sdss psfmag griz
        # use transforms decribed here:
        # https://wiki.sdss.org/display/OPS/All-sky+BOSS+standards#All-skyBOSSstandards-TransformingphotometryofeBOSS-likestandardsintoSDSSsystem  # noqa
        # extract coeffs from fit logs via:
        # gawk 'BEGIN {print("coeffs = {")} /POLYFIT/{printf("\"%s%d\": %s,\n", substr($3,length($3)), $8, $10)} END {print("}")}'  ops_std_eboss/lsdr8_mag_to_sdss_psfmag_*.log  # noqa
        coeffs = {
            "g2": 0.193896,
            "g1": -0.051181,
            "g0": 0.032614,
            "i2": 0.044794,
            "i1": -0.513119,
            "i0": -0.021466,
            "r2": -0.034595,
            "r1": 0.132328,
            "r0": 0.011408,
            "z2": -0.381446,
            "z1": 0.135980,
            "z0": -0.020589,
        }

        g0 = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_g))
        r0 = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_r))
        z0 = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_z))
        g_r = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_g) / peewee.fn.greatest(nMgy_min, ls.flux_r)
        )
        r_z = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r) / peewee.fn.greatest(nMgy_min, ls.flux_z)
        )

        g = g0 + coeffs["g0"] + coeffs["g1"] * g_r + coeffs["g2"] * g_r * g_r
        r = r0 + coeffs["r0"] + coeffs["r1"] * g_r + coeffs["r2"] * g_r * g_r
        i = r0 + coeffs["i0"] + coeffs["i1"] * r_z + coeffs["i2"] * r_z * r_z
        z = z0 + coeffs["z0"] + coeffs["z1"] * r_z + coeffs["z2"] * r_z * r_z

        g0_dered = 22.5 - 2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_g / ls.mw_transmission_g)
        )
        r0_dered = 22.5 - 2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r / ls.mw_transmission_r)
        )
        z0_dered = 22.5 - 2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_z / ls.mw_transmission_z)
        )

        g_r_dered = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_g / ls.mw_transmission_g)
            / peewee.fn.greatest(nMgy_min, ls.flux_r / ls.mw_transmission_r)
        )
        r_z_dered = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r / ls.mw_transmission_r)
            / peewee.fn.greatest(nMgy_min, ls.flux_z / ls.mw_transmission_z)
        )

        # lsdr8 quotes ebv so we can go straight to E(G_BP-G_RP) and E(G_BP-G) directly
        # using the Stassun et al relation:
        # E(BP-RP) = 1.31 * E(B-V)
        # NO! # but we first need to recalibrate the SFD E(B-V) following Schlafly&Finkbeiner2011
        # NO! # such that SF11 E(B-V) = SFD EB(-V) * 0.884
        # NO! E_b_v_corr = 0.884
        # NO! E_bp_rp = 1.31 * E_b_v_corr
        E_bp_rp = 1.31
        R_gaia_g = 1.890 * E_bp_rp
        R_gaia_bp = 2.429 * E_bp_rp
        E_bp_g = R_gaia_bp - R_gaia_g  # = 0.7061

        bp_rp_dered = ls.gaia_phot_bp_mean_mag - ls.gaia_phot_rp_mean_mag - ls.ebv * E_bp_rp
        bp_g_dered = ls.gaia_phot_bp_mean_mag - ls.gaia_phot_g_mean_mag - ls.ebv * E_bp_g

        # bp_rp_dered = (
        #     (
        #         ls.gaia_phot_bp_mean_mag
        #         + 2.5 * peewee.fn.log(ls.mw_transmission_g)
        #     ) - (
        #         ls.gaia_phot_rp_mean_mag
        #         + 2.5 * peewee.fn.log(0.5 * (ls.mw_transmission_r + ls.mw_transmission_z))
        #     )
        # )
        # bp_g_dered = (
        #     (
        #         ls.gaia_phot_bp_mean_mag
        #         + 2.5 * peewee.fn.log(ls.mw_transmission_g)
        #     ) - (
        #         ls.gaia_phot_g_mean_mag
        #         + 2.5 * peewee.fn.log(ls.mw_transmission_r)
        #     )
        # )

        g_r_dered_nominal = pars["g_r_dered_nominal"]
        r_z_dered_nominal = pars["r_z_dered_nominal"]
        bp_rp_dered_nominal = pars["bp_rp_dered_nominal"]
        bp_g_dered_nominal = pars["bp_g_dered_nominal"]

        dered_dist2 = (
            (g_r_dered - g_r_dered_nominal) * (g_r_dered - g_r_dered_nominal)
            + (r_z_dered - r_z_dered_nominal) * (r_z_dered - r_z_dered_nominal)
            + (bp_rp_dered - bp_rp_dered_nominal) * (bp_rp_dered - bp_rp_dered_nominal)
            + (bp_g_dered - bp_g_dered_nominal) * (bp_g_dered - bp_g_dered_nominal)
        )
        dered_dist_max2 = pars["dered_dist_max"] * pars["dered_dist_max"]

        optical_prov = peewee.Value("sdss_psfmag_from_lsdr8")

        query = (
            Catalog.select(
                Catalog.catalogid,
                Catalog.ra,
                Catalog.dec,
                ls.ls_id,
                # ls.flux_g,
                # ls.flux_r,
                # ls.flux_z,
                # ls.flux_w1,
                # ls.flux_ivar_g,
                # ls.flux_ivar_r,
                # ls.flux_ivar_z,
                # ls.flux_ivar_w1,
                g0.alias("ls8_mag_g"),
                r0.alias("ls8_mag_r"),
                z0.alias("ls8_mag_z"),
                g_r.alias("ls8_mag_g_r"),
                r_z.alias("ls8_mag_r_z"),
                g0_dered.alias("ls8_mag_dered_g"),
                r0_dered.alias("ls8_mag_dered_r"),
                z0_dered.alias("ls8_mag_dered_z"),
                g_r_dered.alias("ls8_mag_dered_g_r"),
                r_z_dered.alias("ls8_mag_dered_r_z"),
                bp_rp_dered.alias("gdr2_mag_dered_bp_rp"),
                bp_g_dered.alias("gdr2_mag_dered_bp_g"),
                dered_dist2.alias("dered_dist2"),
                optical_prov.alias("optical_prov"),
                g.alias("g"),
                r.alias("r"),
                i.alias("i"),
                z.alias("z"),
                ls.gaia_phot_g_mean_mag.alias("gaia_g"),
                ls.gaia_phot_bp_mean_mag.alias("bp"),
                ls.gaia_phot_rp_mean_mag.alias("rp"),
                ls.ebv.alias("ls8_ebv"),
                ls.mw_transmission_g.alias("ls8_mw_transmission_g"),
                ls.mw_transmission_r.alias("ls8_mw_transmission_r"),
                ls.mw_transmission_z.alias("ls8_mw_transmission_z"),
                ls.parallax,
                ls.parallax_ivar,
                ls.nobs_g.alias("ls8_nobs_g"),
                ls.nobs_r.alias("ls8_nobs_r"),
                ls.nobs_z.alias("ls8_nobs_z"),
                # ls.maskbits,
            )
            .join(c2ls, on=(Catalog.catalogid == c2ls.catalogid))
            .join(ls, on=(c2ls.target_id == ls.ls_id))
            .where(
                c2ls.version_id == version_id,
                c2ls.best >> True,
                ls.type == "PSF",
                ls.ref_cat == "G2",
                ls.gaia_phot_g_mean_mag > pars["mag_gaia_g_min"],
                ls.parallax < pars["parallax_max"],
                ls.parallax
                > (
                    pars["parallax_min_at_g16"]
                    + (ls.gaia_phot_g_mean_mag - 16.0) * pars["parallax_min_slope"]
                ),
                ls.gaia_duplicated_source >> False,
                ls.nobs_g >= 1,  # TODO increase to >= 2 in lsdr9
                ls.nobs_r >= 1,  # TODO increase to >= 2 in lsdr9
                ls.nobs_z >= 1,  # TODO increase to >= 2 in lsdr9
                ls.flux_g > nMgy_min,
                ls.flux_r > nMgy_min,
                ls.flux_z > nMgy_min,
                ls.maskbits == 0,
                dered_dist2 < dered_dist_max2,
                r0.between(pars["mag_ls_r_min"], pars["mag_ls_r_max"]),
                #
                # g_r.between(pars['ls_g_r_min'], pars['ls_g_r_max']),
                # r_z.between(pars['ls_r_z_min'], pars['ls_r_z_max']),
                # (ls.gaia_phot_bp_mean_mag - ls.gaia_phot_rp_mean_mag)
                # .between(pars['gaia_bp_rp_min'], pars['gaia_bp_rp_max']),
                # (ls.gaia_phot_g_mean_mag - r0)
                # .between(pars['gaia_g_ls_r_min'], pars['gaia_g_ls_r_max'])
                # # (22.5 -
                # #  2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_r)))  # noqa: E501
                # # .between(pars['mag_ls_r_min'], pars['mag_ls_r_max']),
                # # (-2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_g) /  # noqa: E501
                # #                       peewee.fn.greatest(nMgy_min, ls.flux_r)))  # noqa: E501
                # # .between(pars['ls_g_r_min'], pars['ls_g_r_max']),
                # # (-2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_r) /  # noqa: E501
                # #                       peewee.fn.greatest(nMgy_min, ls.flux_z)))  # noqa: E501
                # # .between(pars['ls_r_z_min'], pars['ls_r_z_max']),
                # # (ls.gaia_phot_g_mean_mag -
                # #  (22.5 -
                # #   2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_r))))  # noqa: E501
                # # .between(pars['gaia_g_ls_r_min'], pars['gaia_g_ls_r_max'])
            )
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                ),
                peewee.fn.q3c_radial_query(
                    ls.ra, ls.dec, query_region[0], query_region[1], query_region[2]
                ),
            )
        return query


# ############################################################################


class OPS_BOSS_Stds_LSDR10_Carton(BaseCarton):
    """
    Shorthand name: ops_boss_stds_lsdr10

    Comments: Spectrophotometric standards suitable for use by BOSS in dark time.

              Currently this is a copy of OPS_BOSS_Stds_LSDR8_Carton() with only name changes

    Lead contact:  Tom Dwelly
    """

    name = "ops_std_boss_lsdr10"
    category = "standard_boss"
    cadence = None
    program = "ops_std"
    priority = 5350
    mapper = None
    instrument = "BOSS"
    can_offset = False

    def build_query(self, version_id, query_region=None):
        ls = Legacy_Survey_DR10.alias()
        c2ls = CatalogToLegacy_Survey_DR10.alias()

        # an alias to simplify accessing the query parameters:
        pars = self.parameters

        # safety catch to avoid log-of-zero and divide by zero errors.
        # => use a flux in nano-maggies below which we give up
        nMgy_min = 1e-3  # equiv to AB=30

        # transform the legacysurvey grz into sdss psfmag griz
        # use transforms decribed here:
        # https://wiki.sdss.org/display/OPS/All-sky+BOSS+standards#All-skyBOSSstandards-TransformingphotometryofeBOSS-likestandardsintoSDSSsystem  # noqa
        # extract coeffs from fit logs via:
        # gawk 'BEGIN {print("coeffs = {")} /POLYFIT/{printf("\"%s%d\": %s,\n", substr($3,length($3)), $8, $10)} END {print("}")}'  ops_std_eboss/lsdr8_mag_to_sdss_psfmag_*.log  # noqa
        coeffs = {
            "g2": 0.193896,
            "g1": -0.051181,
            "g0": 0.032614,
            "i2": 0.044794,
            "i1": -0.513119,
            "i0": -0.021466,
            "r2": -0.034595,
            "r1": 0.132328,
            "r0": 0.011408,
            "z2": -0.381446,
            "z1": 0.135980,
            "z0": -0.020589,
        }

        g0 = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_g))
        r0 = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_r))
        i0 = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_i))
        z0 = 22.5 - 2.5 * peewee.fn.log(peewee.fn.greatest(nMgy_min, ls.flux_z))
        g_r = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_g) / peewee.fn.greatest(nMgy_min, ls.flux_r)
        )
        r_i = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r) / peewee.fn.greatest(nMgy_min, ls.flux_i)
        )
        i_z = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_i) / peewee.fn.greatest(nMgy_min, ls.flux_z)
        )
        r_z = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r) / peewee.fn.greatest(nMgy_min, ls.flux_z)
        )

        g = g0 + coeffs["g0"] + coeffs["g1"] * g_r + coeffs["g2"] * g_r * g_r
        r = r0 + coeffs["r0"] + coeffs["r1"] * g_r + coeffs["r2"] * g_r * g_r
        i = r0 + coeffs["i0"] + coeffs["i1"] * r_z + coeffs["i2"] * r_z * r_z
        z = z0 + coeffs["z0"] + coeffs["z1"] * r_z + coeffs["z2"] * r_z * r_z

        g0_dered = 22.5 - 2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_g / ls.mw_transmission_g)
        )
        r0_dered = 22.5 - 2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r / ls.mw_transmission_r)
        )
        i0_dered = 22.5 - 2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_i / ls.mw_transmission_i)
        )
        z0_dered = 22.5 - 2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_z / ls.mw_transmission_z)
        )

        g_r_dered = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_g / ls.mw_transmission_g)
            / peewee.fn.greatest(nMgy_min, ls.flux_r / ls.mw_transmission_r)
        )
        r_i_dered = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r / ls.mw_transmission_r)
            / peewee.fn.greatest(nMgy_min, ls.flux_i / ls.mw_transmission_i)
        )
        i_z_dered = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_i / ls.mw_transmission_i)
            / peewee.fn.greatest(nMgy_min, ls.flux_z / ls.mw_transmission_z)
        )
        r_z_dered = -2.5 * peewee.fn.log(
            peewee.fn.greatest(nMgy_min, ls.flux_r / ls.mw_transmission_r)
            / peewee.fn.greatest(nMgy_min, ls.flux_z / ls.mw_transmission_z)
        )

        # lsdr8 quotes ebv so we can go straight to E(G_BP-G_RP) and E(G_BP-G) directly
        # using the Stassun et al relation:
        # E(BP-RP) = 1.31 * E(B-V)
        # NO! # but we first need to recalibrate the SFD E(B-V) following Schlafly&Finkbeiner2011
        # NO! # such that SF11 E(B-V) = SFD EB(-V) * 0.884
        # NO! E_b_v_corr = 0.884
        # NO! E_bp_rp = 1.31 * E_b_v_corr
        E_bp_rp = 1.31
        R_gaia_g = 1.890 * E_bp_rp
        R_gaia_bp = 2.429 * E_bp_rp
        E_bp_g = R_gaia_bp - R_gaia_g  # = 0.7061

        bp_rp_dered = ls.gaia_phot_bp_mean_mag - ls.gaia_phot_rp_mean_mag - ls.ebv * E_bp_rp
        bp_g_dered = ls.gaia_phot_bp_mean_mag - ls.gaia_phot_g_mean_mag - ls.ebv * E_bp_g

        g_r_dered_nominal = pars["g_r_dered_nominal"]
        r_z_dered_nominal = pars["r_z_dered_nominal"]
        bp_rp_dered_nominal = pars["bp_rp_dered_nominal"]
        bp_g_dered_nominal = pars["bp_g_dered_nominal"]

        dered_dist2 = (
            (g_r_dered - g_r_dered_nominal) * (g_r_dered - g_r_dered_nominal)
            + (r_z_dered - r_z_dered_nominal) * (r_z_dered - r_z_dered_nominal)
            + (bp_rp_dered - bp_rp_dered_nominal) * (bp_rp_dered - bp_rp_dered_nominal)
            + (bp_g_dered - bp_g_dered_nominal) * (bp_g_dered - bp_g_dered_nominal)
        )
        dered_dist_max2 = pars["dered_dist_max"] * pars["dered_dist_max"]

        optical_prov = peewee.Value("sdss_psfmag_from_lsdr8")

        query = (
            Catalog.select(
                Catalog.catalogid,
                Catalog.ra,
                Catalog.dec,
                ls.ls_id,
                g0.alias("ls10_mag_g"),
                r0.alias("ls10_mag_r"),
                i0.alias("ls10_mag_i"),
                z0.alias("ls10_mag_z"),
                g_r.alias("ls10_mag_g_r"),
                r_i.alias("ls10_mag_r_i"),
                i_z.alias("ls10_mag_i_z"),
                r_z.alias("ls10_mag_r_z"),
                g0_dered.alias("ls10_mag_dered_g"),
                r0_dered.alias("ls10_mag_dered_r"),
                i0_dered.alias("ls10_mag_dered_i"),
                z0_dered.alias("ls10_mag_dered_z"),
                g_r_dered.alias("ls10_mag_dered_g_r"),
                r_i_dered.alias("ls10_mag_dered_r_i"),
                i_z_dered.alias("ls10_mag_dered_i_z"),
                r_z_dered.alias("ls10_mag_dered_r_z"),
                bp_rp_dered.alias("gaia_mag_dered_bp_rp"),
                bp_g_dered.alias("gaia_mag_dered_bp_g"),
                dered_dist2.alias("dered_dist2"),
                optical_prov.alias("optical_prov"),
                g.alias("g"),
                r.alias("r"),
                i.alias("i"),
                z.alias("z"),
                ls.gaia_phot_g_mean_mag.alias("gaia_g"),
                ls.gaia_phot_bp_mean_mag.alias("bp"),
                ls.gaia_phot_rp_mean_mag.alias("rp"),
                ls.ebv.alias("ls10_ebv"),
                ls.mw_transmission_g.alias("ls10_mw_transmission_g"),
                ls.mw_transmission_r.alias("ls10_mw_transmission_r"),
                ls.mw_transmission_i.alias("ls10_mw_transmission_i"),
                ls.mw_transmission_z.alias("ls10_mw_transmission_z"),
                ls.parallax,
                ls.parallax_ivar,
                ls.nobs_g.alias("ls10_nobs_g"),
                ls.nobs_r.alias("ls10_nobs_r"),
                ls.nobs_i.alias("ls10_nobs_i"),
                ls.nobs_z.alias("ls10_nobs_z"),
                ls.ref_cat.alias("ls10_ref_cat"),
                ls.ref_id.alias("ls10_ref_id"),
                # ls.maskbits,
            )
            .join(c2ls, on=(Catalog.catalogid == c2ls.catalogid))
            .join(ls, on=(c2ls.target_id == ls.ls_id))
            .where(
                c2ls.version_id == version_id,
                c2ls.best >> True,
                ls.type == "PSF",
                ((ls.ref_cat == "G2") | (ls.ref_cat == "GE")),
                ls.gaia_phot_g_mean_mag > pars["mag_gaia_g_min"],
                ls.parallax < pars["parallax_max"],
                ls.parallax
                > (
                    pars["parallax_min_at_g16"]
                    + (ls.gaia_phot_g_mean_mag - 16.0) * pars["parallax_min_slope"]
                ),
                ls.gaia_duplicated_source >> False,
                ls.nobs_g >= 1,  # TODO increase to >= 2 in lsdr9/10?
                ls.nobs_r >= 1,  # TODO increase to >= 2 in lsdr9/10?
                ls.nobs_z >= 1,  # TODO increase to >= 2 in lsdr9/10?
                ls.flux_g > nMgy_min,
                ls.flux_r > nMgy_min,
                ls.flux_z > nMgy_min,
                ls.maskbits == 0,
                dered_dist2 < dered_dist_max2,
                r0.between(pars["mag_ls_r_min"], pars["mag_ls_r_max"]),
            )
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                ),
                peewee.fn.q3c_radial_query(
                    ls.ra, ls.dec, query_region[0], query_region[1], query_region[2]
                ),
            )
        return query


# ############################################################################


# ############################################################################
class OPS_BOSS_Stds_PS1DR2_Carton(BaseCarton):
    """
    Shorthand name: ops_boss_stds_ps1dr2

    Comments: Spectrophotometric standards suitable for use by BOSS in dark time,
          selected from panstarrs1-dr2 + gaia-dr2.

    Lead contact:  Tom Dwelly
    """

    name = "ops_std_boss_ps1dr2"
    category = "standard_boss"
    cadence = None
    program = "ops_std"
    priority = 5351
    mapper = None
    instrument = "BOSS"
    can_offset = False

    def build_query(self, version_id, query_region=None):
        ps = Panstarrs1.alias()
        c2ps = CatalogToPanstarrs1.alias()
        tic = TIC_v8.alias()
        c2tic = CatalogToTIC_v8.alias()

        # an alias to simplify accessing the query parameters:
        pars = self.parameters

        # transform the panstarrs1-dr2 griz into sdss psfmag griz
        # use transforms decribed here:
        # https://wiki.sdss.org/display/OPS/All-sky+BOSS+standards#All-skyBOSSstandards-TransformingphotometryofeBOSS-likestandardsintoSDSSsystem  # noqa
        # extract coeffs from fit logs via:
        # awk 'BEGIN {print("coeffs = {")} /POLYFIT/{ pe=""; printf("\"%s%d_%s\": %s,\n", substr($3,length($3)), $8, pe, $10)} END {print("}")}'  ops_std_eboss/ps1dr2_chp_psf_to_sdss_psfmag_?_results.log  # noqa

        coeffs = {
            "g2": 0.115563,
            "g1": 0.068765,
            "g0": 0.012047,
            "i2": -0.385214,
            "i1": 0.149677,
            "i0": -0.026127,
            "r2": -0.070151,
            "r1": 0.070129,
            "r0": -0.014197,
            "z2": -2.141255,
            "z1": 0.147746,
            "z0": -0.034845,
        }

        # start from ps1dr2 chp psf mags
        g_r = ps.g_chp_psf - ps.r_chp_psf
        r_i = ps.r_chp_psf - ps.i_chp_psf
        i_z = ps.i_chp_psf - ps.z_chp_psf

        # compute apparent sdss psfmags
        g = ps.g_chp_psf + coeffs["g0"] + coeffs["g1"] * g_r + coeffs["g2"] * g_r * g_r
        r = ps.r_chp_psf + coeffs["r0"] + coeffs["r1"] * g_r + coeffs["r2"] * g_r * g_r
        i = ps.i_chp_psf + coeffs["i0"] + coeffs["i1"] * r_i + coeffs["i2"] * r_i * r_i
        z = ps.z_chp_psf + coeffs["z0"] + coeffs["z1"] * i_z + coeffs["z2"] * i_z * i_z

        # dereddining steps

        # extinction terms for Panstarrs
        # Use R_b from Schlafly & Finkbeiner 2011, Table 6
        # https://ui.adsabs.harvard.edu/abs/2011ApJ...737..103S/abstract
        # assume R_V = 3.1
        # R_b = A_V / E(B-V)
        # for Panstarrs1 grizy bands we have:
        R_g = 3.172
        R_r = 2.271
        R_i = 1.682
        R_z = 1.322
        # R_y = 1.087
        E_g_r = R_g - R_r
        E_r_i = R_r - R_i
        E_i_z = R_i - R_z

        # extinction terms for Gaia
        # Use table 3 from Wang & Chen 2019
        # https://ui.adsabs.harvard.edu/abs/2019ApJ...877..116W/abstract
        # R_G = 1.890
        # R_BP = 2.429
        # R_RP = 1.429
        # These R_b are expressed with E(BP-RP) as the denominator - so need to convert to
        # using E(B-V) as denominator instead

        # an email from Keivan Stassun helps with this:
        #   Hi Jennifer, we had to work
        #   these out for the TIC paper because we used Gaia G mags and
        #   Bp-Rp colors for everything... you can see the details in
        #   Section 2.3.3 of Stassun et al (2019), but here's the final
        #   relations we adopted:
        #   E(${G}_{\mathrm{BP}}$ − ${G}_{\mathrm{RP}}$) = 1.31 E(B − V)
        #   AG = 2.72 E(B − V)

        E_bp_rp = 1.31
        R_gaia_g = 1.890 * E_bp_rp
        R_gaia_bp = 2.429 * E_bp_rp
        # R_gaia_rp = 1.429 * 1.31
        E_bp_g = R_gaia_bp - R_gaia_g  # = 0.7061

        # use ebv from tic_v8 match
        g_r_dered = g_r - tic.ebv * E_g_r
        r_i_dered = r_i - tic.ebv * E_r_i
        i_z_dered = i_z - tic.ebv * E_i_z
        bp_rp_dered = tic.gaiabp - tic.gaiarp - tic.ebv * E_bp_rp
        bp_g_dered = tic.gaiabp - tic.gaiamag - tic.ebv * E_bp_g

        g_r_dered_nominal = pars["g_r_dered_nominal"]
        r_i_dered_nominal = pars["r_i_dered_nominal"]
        i_z_dered_nominal = pars["i_z_dered_nominal"]
        bp_rp_dered_nominal = pars["bp_rp_dered_nominal"]
        bp_g_dered_nominal = pars["bp_g_dered_nominal"]

        dered_dist2 = (
            (g_r_dered - g_r_dered_nominal) * (g_r_dered - g_r_dered_nominal)
            + (r_i_dered - r_i_dered_nominal) * (r_i_dered - r_i_dered_nominal)
            + (i_z_dered - i_z_dered_nominal) * (i_z_dered - i_z_dered_nominal)
            + (bp_rp_dered - bp_rp_dered_nominal) * (bp_rp_dered - bp_rp_dered_nominal)
            + (bp_g_dered - bp_g_dered_nominal) * (bp_g_dered - bp_g_dered_nominal)
        )

        optical_prov = peewee.Value("sdss_psfmag_from_ps1dr2")

        ext_flags = 8388608 + 16777216
        dered_dist_max2 = pars["dered_dist_max"] * pars["dered_dist_max"]

        # the following are just to bracket
        # the result to make the query run faster
        r_stk_psf_flux_min = AB2Jy(pars["mag_ps_r_max"] + 0.2)
        r_stk_psf_flux_max = AB2Jy(pars["mag_ps_r_min"] - 0.2)

        query = (
            c2ps.select(
                c2ps.catalogid,
                tic.ra,
                tic.dec,
                ps.catid_objid.alias("ps1_catid_objid"),
                tic.gaia_int.alias("gaia_source"),
                ps.g_chp_psf.alias("ps1dr2_chp_psfmag_g"),
                ps.r_chp_psf.alias("ps1dr2_chp_psfmag_r"),
                ps.i_chp_psf.alias("ps1dr2_chp_psfmag_i"),
                ps.z_chp_psf.alias("ps1dr2_chp_psfmag_z"),
                g_r.alias("ps1dr2_chp_psfmag_g_r"),
                r_i.alias("ps1dr2_chp_psfmag_r_i"),
                i_z.alias("ps1dr2_chp_psfmag_i_z"),
                tic.ebv.alias("tic_ebv"),
                g_r_dered.alias("ps1dr2_chp_psfmag_g_r_dered"),
                r_i_dered.alias("ps1dr2_chp_psfmag_r_i_dered"),
                i_z_dered.alias("ps1dr2_chp_psfmag_i_z_dered"),
                bp_rp_dered.alias("gdr2_mag_dered_bp_rp"),
                bp_g_dered.alias("gdr2_mag_dered_bp_g"),
                dered_dist2.alias("dered_dist2"),
                optical_prov.alias("optical_prov"),
                g.alias("g"),
                r.alias("r"),
                i.alias("i"),
                z.alias("z"),
                tic.gaiamag.alias("gaia_g"),
                tic.gaiabp.alias("bp"),
                tic.gaiarp.alias("rp"),
                tic.jmag.alias("j"),
                tic.hmag.alias("h"),
                tic.kmag.alias("k"),
                tic.plx.alias("parallax"),
                tic.e_plx.alias("parallax_error"),
            )
            .join(ps, on=(c2ps.target_id == ps.catid_objid))
            .switch(c2ps)
            .join(c2tic, on=(c2ps.catalogid == c2tic.catalogid))
            .join(tic, on=(c2tic.target_id == tic.id))
            .where(
                c2ps.version_id == version_id,
                c2tic.version_id == version_id,
                c2ps.best >> True,
                c2tic.best >> True,
                ps.flags.bin_and(ext_flags) == 0,
                tic.plx < pars["parallax_max"],
                tic.plx
                > (
                    pars["parallax_min_at_g16"] + (tic.gaiamag - 16.0) * pars["parallax_min_slope"]
                ),
                dered_dist2 < dered_dist_max2,
                ps.r_chp_psf.between(pars["mag_ps_r_min"], pars["mag_ps_r_max"]),
                # the following are just attempts to bracket
                # the result to make the query run faster
                tic.gaiamag.between(pars["mag_gaia_g_min"], pars["mag_gaia_g_max"]),
                ps.r_stk_psf_flux.between(r_stk_psf_flux_min, r_stk_psf_flux_max),
                tic.ebv < pars["tic_ebv_max"],
                bp_rp_dered.between(pars["bp_rp_dered_min"], pars["bp_rp_dered_max"]),
            )
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.join_from(CatalogToPanstarrs1, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


# ############################################################################
class OPS_BOSS_Stds_GaiaDR2_Carton(BaseCarton):
    """
    Shorthand name: ops_boss_stds_gdr2

    Comments: Spectrophotometric standards suitable for use by BOSS in dark time,
              selected from tic+gaia-dr2.

    Lead contact:  Tom Dwelly
    """

    name = "ops_std_boss_gdr2"
    category = "standard_boss"
    cadence = None
    program = "ops_std"
    priority = 5352
    mapper = None
    instrument = "BOSS"
    can_offset = False

    def build_query(self, version_id, query_region=None):
        tic = TIC_v8.alias()
        c2tic = CatalogToTIC_v8.alias()

        # an alias to simplify accessing the query parameters:
        pars = self.parameters

        # transform the gaia g,bp,rp into sdss psfmag griz
        # use transforms decribed here:
        # https://wiki.sdss.org/display/OPS/All-sky+BOSS+standards#All-skyBOSSstandards-TransformingphotometryofeBOSS-likestandardsintoSDSSsystem  # noqa
        # extract coeffs from fit logs via:
        # awk 'BEGIN {print("coeffs = {")} /POLYFIT/{ pe=""; printf("\"%s%d%s\": %s,\n", substr($3,length($3)), $8, pe, $10)} END {print("}")}'  ops_std_eboss/gdr2_mag_to_sdss_psfmag_?_results.log  # noqa

        coeffs = {
            "g2": 0.226514,
            "g1": 0.373358,
            "g0": -0.073834,
            "i2": 0.038586,
            "i1": -0.505039,
            "i0": 0.216803,
            "r2": 0.212874,
            "r1": -0.381950,
            "r0": 0.156923,
            "z2": -0.246274,
            "z1": -0.372790,
            "z0": 0.235517,
        }

        bp_rp = tic.gaiabp - tic.gaiarp
        # compute apparent sdss psfmags
        g = tic.gaiamag + coeffs["g0"] + coeffs["g1"] * bp_rp + coeffs["g2"] * bp_rp * bp_rp
        r = tic.gaiamag + coeffs["r0"] + coeffs["r1"] * bp_rp + coeffs["r2"] * bp_rp * bp_rp
        i = tic.gaiamag + coeffs["i0"] + coeffs["i1"] * bp_rp + coeffs["i2"] * bp_rp * bp_rp
        z = tic.gaiamag + coeffs["z0"] + coeffs["z1"] * bp_rp + coeffs["z2"] * bp_rp * bp_rp

        # dereddining steps

        # extinction terms for Gaia
        # Use Stassun+19 (Ticv8) presciption
        E_bp_rp = 1.31
        R_gaia_g = 1.890 * E_bp_rp
        R_gaia_bp = 2.429 * E_bp_rp
        R_gaia_rp = 1.429 * E_bp_rp
        E_bp_g = R_gaia_bp - R_gaia_g  # = 0.7061
        E_g_rp = R_gaia_g - R_gaia_rp  # = 0.6039

        # use ebv from tic_v8 match
        bp_rp_dered = tic.gaiabp - tic.gaiarp - tic.ebv * E_bp_rp
        bp_g_dered = tic.gaiabp - tic.gaiamag - tic.ebv * E_bp_g
        g_rp_dered = tic.gaiamag - tic.gaiarp - tic.ebv * E_g_rp

        bp_rp_dered_nominal = pars["bp_rp_dered_nominal"]
        bp_g_dered_nominal = pars["bp_g_dered_nominal"]
        g_rp_dered_nominal = pars["g_rp_dered_nominal"]

        dered_dist2 = (
            (bp_rp_dered - bp_rp_dered_nominal) * (bp_rp_dered - bp_rp_dered_nominal)
            + (bp_g_dered - bp_g_dered_nominal) * (bp_g_dered - bp_g_dered_nominal)
            + (g_rp_dered - g_rp_dered_nominal) * (g_rp_dered - g_rp_dered_nominal)
        )

        optical_prov = peewee.Value("sdss_psfmag_from_gaia")

        dered_dist_max2 = pars["dered_dist_max"] * pars["dered_dist_max"]

        query = (
            Catalog.select(
                Catalog.catalogid,
                Catalog.ra,
                Catalog.dec,
                tic.id.alias("tic_id"),  # extra
                tic.gaia_int.alias("gaia_source"),  # extra
                tic.ebv.alias("tic_ebv"),  # extra
                bp_rp_dered.alias("gdr2_mag_dered_bp_rp"),  # extra
                bp_g_dered.alias("gdr2_mag_dered_bp_g"),  # extra
                g_rp_dered.alias("gdr2_mag_dered_g_rp"),  # extra
                dered_dist2.alias("dered_dist2"),  # extra
                optical_prov.alias("optical_prov"),
                g.alias("g"),
                r.alias("r"),
                i.alias("i"),
                z.alias("z"),
                tic.gaiamag.alias("gaia_g"),
                tic.gaiabp.alias("bp"),
                tic.gaiarp.alias("rp"),
                tic.jmag.alias("j"),
                tic.hmag.alias("h"),
                tic.kmag.alias("k"),
                tic.plx.alias("parallax"),  # extra
                tic.e_plx.alias("parallax_error"),  # extra
                tic.gallong.alias("tic_gal_l"),  # extra
                tic.gallat.alias("tic_gal_b"),  # extra
            )
            .join(c2tic, on=(Catalog.catalogid == c2tic.catalogid))
            .join(tic, on=(c2tic.target_id == tic.id))
            .where(
                c2tic.version_id == version_id,
                c2tic.best >> True,
                tic.plx < pars["parallax_max"],
                tic.plx
                > (
                    pars["parallax_min_at_g16"] + (tic.gaiamag - 16.0) * pars["parallax_min_slope"]
                ),
                dered_dist2 < dered_dist_max2,
                tic.gaiamag.between(pars["mag_gaia_g_min"], pars["mag_gaia_g_max"]),
                # the following are just to bracket the result to make the query run faster
                tic.gaiabp.between(pars["mag_gaia_bp_min"], pars["mag_gaia_bp_max"]),
                tic.gaiarp.between(pars["mag_gaia_rp_min"], pars["mag_gaia_rp_max"]),
                tic.ebv < pars["ebv_max"],
                ~(tic.gallat.between(-10.0, 10.0)),
            )
        )

        # Below ra, dec and radius are in degrees
        # query_region[0] is ra of center of the region
        # query_region[1] is dec of center of the region
        # query_region[2] is radius of the region
        if query_region:
            query = query.where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                ),
            )

        return query



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2024-07-03
# @Filename: mwm_yso.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTwoMassPSC,
    Gaia_DR3,
    TwoMassPSC,
)

from target_selection.cartons import BaseCarton


# from target_selection.exceptions import TargetSelectionError


# See catalog.py for the name of peewee model names corresponding
# to postgres table names:
# https://github.com/sdss/sdssdb/blob/master/python/sdssdb/peewee/sdss5db/catalogdb.py


class Openfibertargets_mwm_more_ob_boss_Carton(BaseCarton):
    """openfibertargets_mwm_more_ob_boss
    Shorthand name: openfibertargets_mwm_more_ob_boss
    Link: More B stars
    Simplified Description of selection criteria:

    SELECT g.source_id
    FROM gaiadr3.gaia_source as g
    USING (source_id)
    INNER JOIN gaiadr3.tmass_best_neighbour AS xmatch
    ON g.source_id = xmatch.source_id
    INNER JOIN gaiadr1.tmass_original_valid AS tm
    ON tm.tmass_oid = xmatch.tmass_oid

    WHERE parallax < power (10, ((10. - tm.k_m + 0.) / 5.))

    AND parallax > power(10, ((10. - tm.k_m - 0.61) / 5.))

    AND g.phot_g_mean_mag < 16.
    AND tm.j_m - tm.k_m - 0.25 * (g.phot_g_mean_mag - tm.k_m) < 0.10
    AND tm.j_m - tm.k_m - 0.25 * (g.phot_g_mean_mag - tm.k_m) > -0.30
    AND tm.j_m - tm.h_m < 0.15 * (g.phot_g_mean_mag - tm.k_m) + 0.05
    AND tm.j_m - tm.h_m > 0.15 * (g.phot_g_mean_mag - tm.k_m) - 0.15
    AND g.phot_g_mean_mag > 2 * (g.phot_g_mean_mag - tm.k_m) + 3.0

    Return columns: Gaia DR3: source_id, ra, dec, parallax, pmra, pmdec,
    phot_g_mean_mag, phot_bp_mean_mag, phot_rp_mean_mag; 2MASS: j_m, h_m, k_m

    Metadata: can_offset=True
    Priority: 6085
    Cadence: bright_1x1
    Instrument: BOSS
    Program: open_fiber

    Lead contact: Eleonora Zari, Jaime Villaseñor

    """

    name = "openfibertargets_mwm_more_ob_boss"
    category = "science"
    instrument = "BOSS"
    cadence = "bright_1x1"
    program = "open_fiber"
    mapper = "MWM"
    priority = 6085
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra.alias("gaia_dr3_ra"),
                Gaia_DR3.dec.alias("gaia_dr3_dec"),
                Gaia_DR3.parallax,
                Gaia_DR3.pmra,
                Gaia_DR3.pmdec,
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag,
                TwoMassPSC.pts_key,
                TwoMassPSC.designation.alias("twomass_psc_designation"),
                TwoMassPSC.j_m,
                TwoMassPSC.h_m,
                TwoMassPSC.k_m,
            )
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .switch(CatalogToGaia_DR3)
            .join(
                CatalogToTwoMassPSC,
                on=(CatalogToGaia_DR3.catalogid == CatalogToTwoMassPSC.catalogid),
            )
            .join(TwoMassPSC, on=(CatalogToTwoMassPSC.target_id == TwoMassPSC.pts_key))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                CatalogToTwoMassPSC.version_id == version_id,
                CatalogToTwoMassPSC.best >> True,
                Gaia_DR3.phot_g_mean_mag < 16,
                Gaia_DR3.parallax < peewee.fn.power(10, ((10.0 - TwoMassPSC.k_m + 0.00) / 5.0)),
                Gaia_DR3.parallax > peewee.fn.power(10, ((10.0 - TwoMassPSC.k_m - 0.61) / 5.0)),
                TwoMassPSC.j_m
                - TwoMassPSC.k_m
                - 0.25 * (Gaia_DR3.phot_g_mean_mag - TwoMassPSC.k_m)
                < 0.10,
                TwoMassPSC.j_m
                - TwoMassPSC.k_m
                - 0.25 * (Gaia_DR3.phot_g_mean_mag - TwoMassPSC.k_m)
                > -0.30,
                TwoMassPSC.j_m - TwoMassPSC.h_m
                < 0.15 * (Gaia_DR3.phot_g_mean_mag - TwoMassPSC.k_m) + 0.05,
                TwoMassPSC.j_m - TwoMassPSC.h_m
                > 0.15 * (Gaia_DR3.phot_g_mean_mag - TwoMassPSC.k_m) - 0.15,
                Gaia_DR3.phot_g_mean_mag > 2 * (Gaia_DR3.phot_g_mean_mag - TwoMassPSC.k_m) + 3.0,
            )
        )

        # There can be cases in which the same catalogid has multiple entries
        # in a catalog_to_x table since the same physical object
        # may match with multiple catalogids.
        # Hence, we have the below code in the above query.
        #                CatalogToGaia_DR3.best >> True,
        #                CatalogToTwoMassPSC.best >> True,

        # Gaia_DR3 peewee model class corresponds to
        # table catalogdb.gaia_dr3_source.

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query
